<?php

namespace FiloBlu\Refilo\Remote\Entity\Provider;

use FiloBlu\Refilo\Remote\Entity\EntityProviderInterface;
use LogicException;
use Magento\Catalog\Api\Data\ProductInterface;
use Magento\Catalog\Model\ResourceModel\Eav\Attribute;
use Magento\Eav\Api\AttributeRepositoryInterface;
use Magento\Framework\Api\SearchCriteriaBuilder;
use Magento\Store\Api\Data\StoreInterface;
use Magento\Store\Api\Data\WebsiteInterface;
use Magento\Store\Model\StoreManagerInterface;
use Magento\Swatches\Helper\Data;
use Magento\Swatches\Helper\Media;
use Magento\Swatches\Model\Swatch;

/**
 * Class AttributeProvider
 * @package FiloBlu\Refilo\Remote\Entity\Provider
 */
class ProductAttributeProvider implements EntityProviderInterface
{
    /** @var string[] */
    const MANDATORY_ATTRIBUTES = [
        'sku',
        'name',
        'description'
    ];
    /**
     * TODO: attribute to exclude must be configurable
     * @var string[]
     */
    const EXCLUDED_ATTRIBUTES = [
        'allow_message',
        'allow_open_amount',
        'created_at',
        'custom_design',
        'custom_design_from',
        'custom_design_to',
        'custom_layout',
        'custom_layout_update',
        'custom_layout_update_file',
        'email_template',
        'has_options',
        'media_gallery',
        'minimal_price',
        'price',
        'msrp_display_actual_price_type',
        'old_id',
        'open_amount_max',
        'open_amount_min',
        'image',
        'image_label',
        'links_exist',
        'links_purchased_separately',
        'links_title',
        'merchant_center_category',
        'msrp',
        'open_amount_max',
        'open_amount_min',
        'options_container',
        'origin_storage',
        'package_mode',
        'package_type',
        'page_layout',
        'price_type',
        'price_view',
        'quantity_and_stock_status',
        'related_tgtr_position_behavior',
        'related_tgtr_position_limit',
        'required_options',
        'small_image',
        'small_image_label',
        'special_price',
        'status',
        'swatch_image',
        'thumbnail',
        'thumbnail_label',
        'tier_price',
        'type_of_product',
        'updated_at',
        'upsell_tgtr_position_behavior',
        'upsell_tgtr_position_limit',
        'use_config_allow_message',
        'use_config_email_template',
        'use_config_is_redeemable',
        'visibility',
        'shipment_type'
    ];
    /** @var StoreInterface */
    protected $currentStore = null;
    /** @var array|null */
    protected $currentAttributes = null;
    /** @var array */
    protected $imagePlaceholderAttributes = [];
    /** @var array */
    protected $attributes;
    /** @var array */
    protected $filterableAttributes = [];
    /** @var array */
    protected $sortableAttributes = [];
    /** @var array */
    protected $lookupOptions = [];
    /** @var SearchCriteriaBuilder */
    private $searchCriteriaBuilder;
    /** @var AttributeRepositoryInterface */
    private $attributeRepository;
    /** @var StoreManagerInterface */
    private $storeManager;
    /** @var Data */
    private $swatchHelper;
    /** @var Media */
    private $swatchMedia;

    /**
     * ProductAttributeProvider constructor.
     * @param SearchCriteriaBuilder $searchCriteriaBuilder
     * @param AttributeRepositoryInterface $attributeRepository
     * @param StoreManagerInterface $storeManager
     * @param Data $swatchHelper
     * @param Media $swatchMedia
     */
    public function __construct(
        SearchCriteriaBuilder $searchCriteriaBuilder,
        AttributeRepositoryInterface $attributeRepository,
        StoreManagerInterface $storeManager,
        Data $swatchHelper,
        Media $swatchMedia
    ) {
        $this->searchCriteriaBuilder = $searchCriteriaBuilder;
        $this->attributeRepository = $attributeRepository;
        $this->storeManager = $storeManager;
        $this->swatchHelper = $swatchHelper;
        $this->swatchMedia = $swatchMedia;
    }

    /**
     * @param ProductInterface $product
     * @return array
     */
    public function getAttributesValues(ProductInterface $product)
    {
        $this->init();
        return $this->getAttributesFromProduct($product, $this->attributes, self::EXCLUDED_ATTRIBUTES);
    }

    /**
     *
     */
    protected function init()
    {
        if ($this->attributes !== null) {
            return;
        }

        $searchCriteria = $this->searchCriteriaBuilder->create();
        $response = $this->attributeRepository->getList('catalog_product', $searchCriteria);

        foreach ($this->storeManager->getStores() as $store) {
            $storeId = $store->getId();
            $this->filterableAttributes[$storeId] = [];
            $this->sortableAttributes[$storeId] = [];
            $this->attributes[$storeId] = [];
        }

        foreach ($this->storeManager->getStores() as $store) {
            $storeId = $store->getId();
            /** @var Attribute $attribute */
            foreach ($response->getItems() as $attribute) {
                $attributeCode = $attribute->getAttributeCode();
                $isSwatch = (bool)$this->swatchHelper->isSwatchAttribute($attribute);

                if ((bool)$attribute->getIsVisibleOnFront() ||
                    (bool)$attribute->getUsedInProductListing() ||
                    $this->isMandatory($attributeCode)) {
                    $this->attributes[$storeId][$attributeCode] = [
                        'title'        => trim($attribute->getStoreLabel($store->getId())),
                        'code'         => $attributeCode,
                        'type'         => $attribute->getFrontendInput(),
                        'isSwatch'     => $isSwatch,
                        'isFilterable' => false,
                        'isSortable'   => false,
                        'searchable'   => $attribute->getIsSearchable(),
                        'search_weight' => $attribute->getSearchWeight(),
                        'sort_weight'   => $attribute->getSortWeight($attribute->getAttributeSetId()),
                        'position'     => $attribute->getPosition(),
                        'options'      => []
                    ];

                    if ($attribute->getIsFilterable()) {
                        $this->attributes[$storeId][$attributeCode]['isFilterable'] = true;
                        $this->filterableAttributes[$storeId][$attributeCode] = $this->attributes[$storeId][$attributeCode];
                    }

                    if ($attribute->getUsedForSortBy()) {
                        $this->attributes[$storeId][$attributeCode]['isSortable'] = true;
                        $this->sortableAttributes[$storeId][$attributeCode] = $this->attributes[$storeId][$attributeCode];
                    }
                }

                if ($attribute->getFrontendInput() === 'media_image') {
                    $this->imagePlaceholderAttributes[$attributeCode] = $attributeCode;
                }

                /* if($attribute->getFrontend()->getConfigField('input') === 'boolean') {} */

                if ($attribute->usesSource()) {
                    foreach ($attribute->getSource()->getAllOptions() as $option) {
                        $this->lookupOptions[$store->getId()][$attributeCode][$option['value']] = (string)$option['label'];
                        if (isset($this->attributes[$storeId][$attributeCode]['options'])) {
                            $optionValue = [
                                'label'  => trim((string)$option['label']),
                                'key'    => trim($option['value']),
                                'swatch' => []
                            ];

                            if ($isSwatch) {
                                $optionValue = array_merge_recursive($optionValue, $this->getSwatchByOptionId($option['value']));
                            }

                            $this->attributes[$storeId][$attributeCode]['options'][] = $optionValue;
                        }
                    }
                }
            }
        }
    }

    /**
     * @param string $attributeCode
     * @return bool
     */
    public function isMandatory(string $attributeCode): bool
    {
        return in_array($attributeCode, self::MANDATORY_ATTRIBUTES, false);
    }

    /**
     * @param $optionId
     * @return array
     */
    public function getSwatchByOptionId($optionId)
    {
        $output = [
            'swatch' => [
                'type'  => 'empty',
                'value' => ''
            ]
        ];
        $swatches = $this->swatchHelper->getSwatchesByOptionsId([$optionId]);
        if (!isset($swatches[$optionId]['type'], $swatches[$optionId]['value'])) {
            return $output;
        }

        $type = (int)$swatches[$optionId]['type'];
        $value = $swatches[$optionId]['value'];

        switch ($type) {
            case Swatch::SWATCH_TYPE_VISUAL_IMAGE:
                // Backward compatibility
                return [
                    'type'   => 'swatch',
                    'image'  => "/attribute/swatch{$value}",
                    'swatch' => [
                        'type'  => 'swatch',
                        'value' => "/attribute/swatch{$value}"
                    ]
                ];
                break;
            case Swatch::SWATCH_TYPE_VISUAL_COLOR:
                return [
                    'swatch' => [
                        'type'  => 'color',
                        'value' => $value
                    ]
                ];
                break;
            case Swatch::SWATCH_TYPE_TEXTUAL:
                return [
                    'swatch' => [
                        'type'  => 'textual',
                        'value' => $value
                    ]
                ];
                break;
            case Swatch::SWATCH_TYPE_EMPTY:
            default:
                return $output;
                break;
        }
    }

    /**
     * @param ProductInterface $product
     * @param array $attributeList
     * @param array $excludeList
     * @return array
     */
    protected function getAttributesFromProduct(ProductInterface $product, array $attributeList, array $excludeList = []): array
    {
        $storeId = $product->getStore()->getId();

        $attributes = [];

        foreach ($attributeList[$storeId] as $code => $attribute) {
            if (in_array($code, $excludeList, true)) {
                continue;
            }

            $type = $attribute['type'];

            switch ($type) {
                case 'select':
                    $attributes[$code] = [
//                        'code'  => $code,
//                        'title' => $attribute['title'],
//                        'type'  => $type,
                        'value' => $this->lookupOptions[$storeId][$code][$product->getData($code)] ?? null,
                        'key'   => $product->getData($code)
                    ];

                    if ($attribute['isSwatch']) {
                        $data = $this->getSwatchByOptionId($product->getData($code));
                        $attributes[$code] = array_merge_recursive($attributes[$code], $data);
                    }
                    break;
                case 'multiselect':

                    $value = $product->getAttributeText($code) ?: [];

                    if (is_scalar($value)) {
                        $value = [$value];
                    }

                    // TODO: getAttributeText is slow, prefetch values
                    $attributes[$code] =
                        [
//                            'code'  => $code,
//                            'title' => $attribute['title'],
//                            'type'  => $type,
                            'value' => $value,
                            'key'   => array_map('intval', array_filter(explode(',', $product->getData($code) ?? ''), 'strlen')),
                        ];
                    break;
                default:
                    $attributes[$code] =
                        [
//                            'code'  => $code,
//                            'title' => $attribute['title'],
//                            'type'  => $type,
                            'value' => $product->getData($code)
                        ];
                    break;
            }
        }

        return $attributes;
    }

    /**
     * @param ProductInterface $product
     * @return array
     */
    public function getFilterableAttributes(ProductInterface $product): array
    {
        $this->init();
        return $this->getAttributesFromProduct($product, $this->filterableAttributes);
    }

    /**
     * @param ProductInterface $product
     * @return array
     */
    public function getSortableAttributes(ProductInterface $product)
    {
        $this->init();
        return $this->getAttributesFromProduct($product, $this->sortableAttributes);
    }

    /**
     * @param ProductInterface $product
     * @return array
     */
    public function getImagePlaceholdersAttributes(ProductInterface $product): array
    {
        $this->init();

        $placeholders = [];

        foreach ($this->imagePlaceholderAttributes as $attributeCode) {
            $value = $product->getData($attributeCode);
            if ($value === null) {
                continue;
            }
            $placeholders[$attributeCode] = "/catalog/product{$value}";
        }

        return $placeholders;
    }

    /**
     * @return mixed
     */
    public function current()
    {
        $attribute = new \FiloBlu\Refilo\Remote\Entity\Attribute();
        $attribute->setId(key($this->currentAttributes));
        $attribute->addData(current($this->currentAttributes));

        return $attribute;
    }

    /**
     * @return mixed|void
     */
    public function next()
    {
        return next($this->currentAttributes);
    }

    /**
     * @return bool|float|int|string|null
     */
    public function key()
    {
        return key($this->currentAttributes);
    }

    /**
     * @return bool
     */
    public function valid()
    {
        return $this->attributes !== null && current($this->currentAttributes);
    }

    public function rewind()
    {
        $this->currentAttributes = $this->getAttributes($this->currentStore);
    }

    /**
     * @param StoreInterface $store
     * @return mixed
     */
    public function getAttributes(StoreInterface $store)
    {
        $this->init();
        return $this->attributes[$store->getId()];
    }

    public function withIds(array $ids): EntityProviderInterface
    {
        return $this;
    }

    public function forStore(StoreInterface $store): EntityProviderInterface
    {
        $this->currentStore = $store;
        return $this;
    }

    public function forWebsite(WebsiteInterface $website): EntityProviderInterface
    {
        throw new LogicException('Count not filter by website');
    }
}
